from typing import Dict, List, Type

import instructor
from openai import OpenAI
from pydantic import BaseModel


def completion(
    messages: List[Dict[str, str]],
    openai_key: str,
    instructor_model: Type[BaseModel],
    max_retries: int = 3,
) -> Type[BaseModel]:
    """
    Most of you are likely familiar with tool calling with GPT4.

    This is just that but instead of passing a JSON string, we
    pass a Pydantic model, and it returns that Pydantic model.

    In additon we can add retry mechanisms. This is all thanks to
    Jason Liu and his instructor package. Definitely recommend
    checking it out!

    https://github.com/jxnl/instructor

    Args:
    - messages: List[Dict[str, str]]: The messages to send to the model
    - openai_key: str: The OpenAI API key
    - model: Type[BaseModel]: The Pydantic model to return
    - max_retries: int: The maximum number of retries to attempt

    Returns:
    - Type[BaseModel]: The Pydantic model
    """
    client = instructor.patch(OpenAI(api_key=openai_key))
    return client.chat.completions.create(
        model="gpt-4-1106-preview",
        messages=messages,
        response_model=instructor_model,
        temperature=0.0,
        seed=69,
        max_retries=max_retries,
    )



"""
The spawn operation will create a new node in the swarm, and begin executing the action assigned to the node.
"""
from importlib import import_module
from typing import List, Union

from swarmstar.swarm.types import (
    ActionSpace,
    SpawnOperation,
    SwarmConfig,
    SwarmHistory,
    SwarmNode,
    SwarmOperation,
    SwarmState,
)
from swarmstar.utils.misc.uuid import generate_uuid


def spawn(
    swarm: SwarmConfig, spawn_operation: SpawnOperation
) -> Union[SwarmOperation, List[SwarmOperation]]:
    swarm_state = SwarmState(swarm=swarm)
    action_space = ActionSpace(swarm=swarm)

    parent_id = spawn_operation.node_id
    node_embryo = spawn_operation.node_embryo
    action_id = node_embryo.action_id

    action_metadata = action_space[action_id]
    termination_policy = action_metadata.termination_policy
    if spawn_operation.termination_policy_change is not None:
        termination_policy = spawn_operation.termination_policy_change
    node = SwarmNode(
        node_id=generate_uuid(action_space[action_id].name),
        parent_id=parent_id,
        action_id=action_id,
        message=node_embryo.message,
        alive=True,
        termination_policy=termination_policy
    )
    swarm_state.update_state(node)

    if parent_id is not None:
        parent_node = swarm_state[parent_id]
        parent_node.children_ids.append(node.node_id)
        swarm_state.update_state(parent_node)

    swarm_history = SwarmHistory(swarm=swarm)
    swarm_history.add_event(spawn_operation)
    output = execute_node_action(swarm, node, action_metadata)
    return output


def execute_node_action(
    swarm: SwarmConfig, node: SwarmNode, action_metadata
) -> Union[SwarmOperation, List[SwarmOperation]]:
    action_type = action_metadata.type

    action_type_map = {
        "internal_action": "swarmstar.utils.swarm_utils.spawn_operations.internal_action",
    }

    if action_type not in action_type_map:
        raise ValueError(
            f"Action type: `{action_type}` from action id: `{node.action_id}` is not supported yet."
        )

    action_type_module = import_module(action_type_map[action_type])
    action_output = action_type_module.execute_action(swarm, node, action_metadata)
    return action_output



"""
Azure Blob Storage doesen't have a concept of folders. It has a flat
structure and uses the '/' character in the blob name to simulate a folder.

For a common interface to folder operations I've still implemented these, but
make_folder() does nothing and rename_folder() and move_folder() are the
same thing.
"""
from __future__ import annotations

from typing import TYPE_CHECKING, List

from azure.storage.blob import BlobServiceClient

if TYPE_CHECKING:
    from swarmstar.swarm.types import SwarmConfig


def delete_folder(swarm: SwarmConfig, folder_path: str) -> dict:
    try:
        blob_service_client = BlobServiceClient(
            account_url=f"https://{swarm.azure_blob_storage_account_name}.blob.core.windows.net/",
            credential=swarm.azure_blob_storage_account_key,
        )
        container_client = blob_service_client.get_container_client(
            swarm.azure_blob_storage_container_name
        )

        blobs_list = list(container_client.list_blobs(name_starts_with=folder_path))
        for blob in blobs_list:
            container_client.delete_blob(blob.name)
    except Exception as e:
        raise ValueError(f"Failed to delete folder: {str(e)}")


def list_folder(swarm: SwarmConfig, folder_path: str) -> List[str]:
    try:
        blob_service_client = BlobServiceClient(
            account_url=f"https://{swarm.azure_blob_storage_account_name}.blob.core.windows.net/",
            credential=swarm.azure_blob_storage_account_key,
        )
        container_client = blob_service_client.get_container_client(
            swarm.azure_blob_storage_container_name
        )

        # Ensure folder_path ends with a '/' to properly simulate a folder structure
        if not folder_path.endswith("/"):
            folder_path += "/"

        blob_list = container_client.list_blobs(name_starts_with=folder_path)
        paths = []
        for blob in blob_list:
            # Split the blob name by '/' and filter out those that are more than one level deep
            if blob.name[len(folder_path) :].count("/") == 0:
                paths.append(blob.name)
    except Exception as e:
        raise ValueError(f"Failed to list folder: {str(e)}")


def make_folder(swarm: SwarmConfig, folder_path: str) -> None:
    return None


def move_folder(swarm: SwarmConfig, folder_path: str, new_folder_name: str) -> None:
    try:
        rename_folder(swarm, folder_path, new_folder_name)
    except Exception as e:
        raise ValueError(f"Failed to move folder: {str(e)}")


def rename_folder(swarm: SwarmConfig, folder_path: str, new_folder_path: str) -> None:
    try:
        blob_service_client = BlobServiceClient(
            account_url=f"https://{swarm.azure_blob_storage_account_name}.blob.core.windows.net/",
            credential=swarm.azure_blob_storage_account_key,
        )
        container_client = blob_service_client.get_container_client(
            swarm.azure_blob_storage_container_name
        )

        blobs = list(container_client.list_blobs(name_starts_with=folder_path))
        for blob in blobs:
            # Construct the new blob name
            new_blob_name = new_folder_path + blob.name[len(folder_path) :]

            # Copy the blob to the new location
            copied_blob = container_client.get_blob_client(new_blob_name)
            copied_blob.start_copy_from_url(blob.url)

            container_client.delete_blob(blob.name)
    except Exception as e:
        raise ValueError(f"Failed to rename folder: {str(e)}")



from __future__ import annotations

from typing import TYPE_CHECKING

from azure.storage.blob import BlobServiceClient

if TYPE_CHECKING:
    from swarmstar.swarm.types import SwarmConfig


def delete_file(swarm: SwarmConfig, file_path: str) -> None:
    storage_account_name = swarm.azure_blob_storage_account_name
    storage_account_key = swarm.azure_blob_storage_account_key
    container_name = swarm.azure_blob_storage_container_name

    try:
        # Create a blob service client
        blob_service_client = BlobServiceClient(
            account_url=f"https://{storage_account_name}.blob.core.windows.net/",
            credential=storage_account_key,
        )

        # Get a client to interact with the specified container
        container_client = blob_service_client.get_container_client(container_name)

        # Delete the blob
        container_client.delete_blob(file_path)
    except Exception as e:
        raise ValueError(f"Failed to delete file: {str(e)}")


def move_file(swarm: SwarmConfig, file_path: str, new_file_path: str) -> None:
    storage_account_name = swarm.azure_blob_storage_account_name
    storage_account_key = swarm.azure_blob_storage_account_key
    container_name = swarm.azure_blob_storage_container_name

    try:
        # Create a blob service client
        blob_service_client = BlobServiceClient(
            account_url=f"https://{storage_account_name}.blob.core.windows.net/",
            credential=storage_account_key,
        )

        # Get a client to interact with the specified container
        container_client = blob_service_client.get_container_client(container_name)

        # Get a blob client to interact with the specified blob
        blob_client = container_client.get_blob_client(file_path)

        # Copy the blob
        blob_client.start_copy_from_url(new_file_path)
    except Exception as e:
        raise ValueError(f"Failed to move file: {str(e)}")


def rename_file(swarm: SwarmConfig, file_path: str, new_file_name: str) -> None:
    storage_account_name = swarm.azure_blob_storage_account_name
    storage_account_key = swarm.azure_blob_storage_account_key
    container_name = swarm.azure_blob_storage_container_name

    try:
        blob_service_client = BlobServiceClient(
            account_url=f"https://{storage_account_name}.blob.core.windows.net/",
            credential=storage_account_key,
        )

        container_client = blob_service_client.get_container_client(container_name)

        blob_client = container_client.get_blob_client(file_path)

        new_blob_client = container_client.get_blob_client(new_file_name)
        new_blob_client.start_copy_from_url(blob_client.url)

        # TODO: In a production environment, consider implementing a more robust solution for monitoring copy status.
        import time

        while new_blob_client.get_blob_properties().copy.status != "success":
            time.sleep(1)

        blob_client.delete_blob()

    except Exception as e:
        raise ValueError(f"Failed to rename file: {str(e)}")


def upload_file(swarm: SwarmConfig, file_path: str, file_bytes: bytes) -> None:
    storage_account_name = swarm.azure_blob_storage_account_name
    storage_account_key = swarm.azure_blob_storage_account_key
    container_name = swarm.azure_blob_storage_container_name

    try:
        # Create a blob service client
        blob_service_client = BlobServiceClient(
            account_url=f"https://{storage_account_name}.blob.core.windows.net/",
            credential=storage_account_key,
        )

        # Get a client to interact with the specified container
        container_client = blob_service_client.get_container_client(container_name)

        # Create the container if it does not exist
        container_client.create_container()

        # Get a blob client to interact with the specified blob
        blob_client = container_client.get_blob_client(file_path)

        # Upload the file
        blob_client.upload_blob(file_bytes, overwrite=True)
    except Exception as e:
        raise ValueError(f"Failed to upload file: {str(e)}")


def retrieve_file(swarm: SwarmConfig, file_path: str) -> bytes:
    storage_account_name = swarm.azure_blob_storage_account_name
    storage_account_key = swarm.azure_blob_storage_account_key
    container_name = swarm.azure_blob_storage_container_name

    try:
        # Create a blob service client
        blob_service_client = BlobServiceClient(
            account_url=f"https://{storage_account_name}.blob.core.windows.net/",
            credential=storage_account_key,
        )

        # Get a client to interact with the specified container
        container_client = blob_service_client.get_container_client(container_name)

        # Get a blob client to interact with the specified blob
        blob_client = container_client.get_blob_client(file_path)

        # Download the blob
        return blob_client.download_blob().readall()

    except Exception as e:
        raise ValueError(f"Failed to retrieve file: {str(e)}")



"""
Use this object to configure the swarm to your platform and personal preferences.
"""

from typing import Optional

from pydantic import BaseModel
from typing_extensions import Literal


class SwarmConfig(BaseModel):
    root_path: str
    openai_key: str
    platform: Literal["mac", "azure"]
    user_id: Optional[str] = None
    swarm_id: Optional[str] = None
    azure_blob_storage_account_name: Optional[str] = None
    azure_blob_storage_account_key: Optional[str] = None
    azure_blob_storage_container_name: Optional[str] = None
    azure_cosmos_db_url: Optional[str] = None
    azure_cosmos_db_key: Optional[str] = None
    azure_cosmos_db_database_name: Optional[str] = None
    azure_cosmos_db_container_name: Optional[str] = None
    mongodb_uri: Optional[str] = None
    mongodb_db_name: Optional[str] = None


# The types below aren't actually used in the codebase, but they are used in the documentation.


class PlatformConfig(BaseModel):
    platform: Literal["mac", "azure"]


class AzureConfig(PlatformConfig):
    platform: Literal["azure"]
    user_id: str
    swarm_id: str
    azure_blob_storage_account_name: str = None
    azure_blob_storage_account_key: str = None
    azure_blob_storage_container_name: str = None
    azure_cosmos_db_url: str = None
    azure_cosmos_db_key: str = None
    azure_cosmos_db_database_name: str = None
    azure_cosmos_db_container_name: str = None


class LocalConfig(PlatformConfig):
    platform: Literal["mac"]
    mongodb_uri: str = None
    mongodb_db_name: str = None



"""
Swarm setup includes:
    1. Setting up storage for the swarm space on your chosen platform
    2. Creating a swarm object
    
The swarm space contains all the metadata for the swarm spaces, a stage
for generated content, a place to store the state of the swarm, and more.
    
The swarm object contains keys and configs for your personal swarm space. It's
passed around between every node and action. Keep it safe and private!
"""
import os

from swarmstar.swarm.types import SwarmConfig
from swarmstar.utils.data.kv_operations.main import add_kv
from swarmstar.utils.data.kv_operations.mongodb import (
    check_and_create_database,
)

def configure_swarm(
    openai_key: str, root_path: str, platform: str, **kwargs
) -> SwarmConfig:
    try:
        if platform not in ["mac", "azure"]:
            raise ValueError
    except ValueError:
        raise ValueError(f"Invalid platform: {platform}")

    platform_map = {
        "mac": _setup_mac_swarm_space,
        "azure": _setup_azure_swarm_space,
    }

    swarm = platform_map[platform](openai_key, root_path, **kwargs)
    add_kv(swarm, "swarm_history", "current_frame", {"frame": 0})
    add_kv(swarm, "swarm", "swarm", swarm.model_dump())
    return swarm


def _setup_mac_swarm_space(openai_key: str, root_path: str, **kwargs) -> SwarmConfig:
    required_keys = ["mongodb_uri", "mongodb_db_name"]
    missing_keys = [key for key in required_keys if key not in kwargs]
    if missing_keys:
        raise ValueError(f'Missing required key(s): {", ".join(missing_keys)}')

    if os.path.exists(root_path):
        if os.listdir(root_path):
            raise ValueError(f"Root path folder must be empty: {root_path}")
    else:
        os.makedirs(root_path)

    check_and_create_database(kwargs["mongodb_uri"], kwargs["mongodb_db_name"])

    return SwarmConfig(
        openai_key=openai_key,
        root_path=root_path,
        platform="mac",
        mongodb_uri=kwargs["mongodb_uri"],
        mongodb_db_name=kwargs["mongodb_db_name"],
    )


def _setup_azure_swarm_space(openai_key: str, root_path: str, **kwargs) -> SwarmConfig:
    # Test connection to Azure Blob Storage
    from azure.storage.blob import BlobServiceClient

    storage_account_name = kwargs["azure_blob_storage_account_name"]
    storage_account_key = kwargs["azure_blob_storage_account_key"]
    container_name = kwargs["azure_blob_storage_container_name"]

    try:
        blob_service_client = BlobServiceClient(
            account_url=f"https://{storage_account_name}.blob.core.windows.net/",
            credential=storage_account_key,
        )
    except Exception as e:
        raise ValueError(f"Failed to connect to Azure Blob Storage: {str(e)}")

    # Test connection to Azure CosmosDB
    from azure.cosmos import CosmosClient

    cosmos_db_url = kwargs["azure_cosmos_db_url"]
    cosmos_db_key = kwargs["azure_cosmos_db_key"]
    database_name = kwargs["azure_cosmos_db_database_name"]
    container_name = kwargs["azure_cosmos_db_container_name"]
    try:
        cosmos_client = CosmosClient(cosmos_db_url, credential=cosmos_db_key)
        database = cosmos_client.get_database_client(database_name)
        database.get_container_client(container_name)
    except Exception as e:
        raise ValueError(f"Failed to connect to Azure CosmosDB: {str(e)}")

    return SwarmConfig(
        openai_key=openai_key,
        root_path=root_path,
        platform="azure",
        user_id=kwargs["user_id"],
        swarm_id=kwargs["swarm_id"],
        azure_blob_storage_account_name=kwargs["azure_blob_storage_account_name"],
        azure_blob_storage_account_key=kwargs["azure_blob_storage_account_key"],
        azure_blob_storage_container_name=kwargs["azure_blob_storage_container_name"],
        azure_cosmos_db_url=kwargs["azure_cosmos_db_url"],
        azure_cosmos_db_key=kwargs["azure_cosmos_db_key"],
        azure_cosmos_db_database_name=kwargs["azure_cosmos_db_database_name"],
        azure_cosmos_db_container_name=kwargs.get("azure_cosmos_db_container_name"),
    )



"""
The swarm state is a record of the current state of the swarm. It
simply stores a dictionary of all nodes. This state can be deconstructed
and reconstructed from the history.
"""
from pydantic import BaseModel

from swarmstar.swarm.types.swarm_config import SwarmConfig
from swarmstar.swarm.types.swarm_nodes import SwarmNode
from swarmstar.utils.data.kv_operations.main import add_kv, get_kv


class SwarmState(BaseModel):
    swarm: "SwarmConfig"

    def __getitem__(self, node_id: str) -> SwarmNode:
        node = SwarmNode.model_validate(get_kv(self.swarm, "swarm_state", node_id))
        return node

    def update_state(self, node: SwarmNode):
        add_kv(self.swarm, "swarm_state", node.node_id, node.model_dump())



from importlib import import_module
from typing import List, Union

from swarmstar.swarm.types import Action, SwarmConfig, SwarmNode, SwarmOperation


def execute_action(
    swarm: SwarmConfig, node: SwarmNode, action_metadata: Action
) -> Union[SwarmOperation, List[SwarmOperation]]:
    internal_action_path = action_metadata.internal_action_path
    action_name = action_metadata.name
    action_class = getattr(import_module(internal_action_path), action_name)
    action_instance = action_class(swarm=swarm, node=node)
    return action_instance.main()



"""
The swarm util space metadata has a simlar but different role to 
the action space metadata:

    (1) To allow the swarm to organize and search for utils
    (2) Describe how to interact with the utils

Swarm utils are just swarm specific functions. They may be
used in the construction of actions.

For now we assume all utils are internal to the package.
"""
from typing import Dict, List

from pydantic import BaseModel
from typing_extensions import Literal

from swarmstar.swarm.types.swarm_config import SwarmConfig
from swarmstar.utils.data.internal_operations import get_internal_util_metadata
from swarmstar.utils.data.kv_operations.main import get_kv


class UtilMetadata(BaseModel):
    type: Literal["internal_folder", "internal_function"]
    name: str
    description: str
    parent: str
    children: List[str] = []
    metadata: Dict[str, str]


class UtilSpace(BaseModel):
    swarm: SwarmConfig

    def __getitem__(self, util_id: str) -> UtilMetadata:
        try:
            internal_util_metadata = get_internal_util_metadata(util_id)
            return internal_util_metadata
        except Exception:
            external_util_metadata = get_kv(self.swarm, "util_space", util_id)
            if external_util_metadata is not None:
                return external_util_metadata
            else:
                raise ValueError(f"This util id: `{util_id}` does not exist.")



"""
This blocking operation will call the next_function_to_call of the node's action,
combining the args, context and swarm into the function call.

Think of this as the entrypoint back into the action after a blocking operation has been completed.
"""

from __future__ import annotations

from importlib import import_module
from typing import TYPE_CHECKING, List, Union

from pydantic import BaseModel

from swarmstar.swarm.types import ActionSpace, SwarmState

if TYPE_CHECKING:
    from swarmstar.swarm.types import BlockingOperation, SwarmConfig, SwarmOperation


# This blocking operation doesn't have set args, it will just pass the args and context to the next function to call
class expected_args(BaseModel):
    any


def blocking(
    swarm: SwarmConfig, blocking_operation: BlockingOperation
) -> Union[SwarmOperation, List[SwarmOperation]]:
    action_space = ActionSpace(swarm=swarm)
    swarm_state = SwarmState(swarm=swarm)
    node_id = blocking_operation.node_id
    node = swarm_state[node_id]
    action_metadata = action_space[node.action_id]
    action_name = action_metadata.name
    script_path = action_metadata.internal_action_path
    action_file = import_module(script_path)
    combined_args = {}
    combined_args.update(blocking_operation.args)
    if blocking_operation.context is not None:
        combined_args.update(blocking_operation.context)

    action_class = getattr(action_file, action_name)
    action_instance = action_class(swarm=swarm, node=node)
    return getattr(action_instance, blocking_operation.next_function_to_call)(
        **combined_args
    )



from importlib import import_module
from typing import List, Union

from swarmstar.swarm.types import (
    BlockingOperation,
    SwarmConfig,
    SwarmHistory,
    SwarmOperation,
)


def blocking(
    swarm: SwarmConfig, blocking_operation: BlockingOperation
) -> Union[SwarmOperation, List[SwarmOperation]]:
    blocking_operation_type_map = {
        "instructor_completion": "swarmstar.utils.swarm_utils.blocking_operations.instructor.completion",
        "internal_action": "swarmstar.utils.swarm_utils.blocking_operations.internal_action",
    }

    blocking_operation_type = blocking_operation.blocking_type

    if blocking_operation_type not in blocking_operation_type_map:
        raise ValueError(
            f"Blocking operation type: `{blocking_operation_type.type}` is not supported."
        )

    blocking_operation_type_module = import_module(
        blocking_operation_type_map[blocking_operation_type]
    )
    output: SwarmOperation = blocking_operation_type_module.blocking(
        swarm, blocking_operation
    )

    swarm_history = SwarmHistory(swarm=swarm)
    swarm_history.add_event(blocking_operation)

    return output



from .swarm_config import SwarmConfig, PlatformConfig, AzureConfig, LocalConfig

from .action_metadata import (
    ActionNode,
    Action,
    ActionFolder,
    InternalAction,
    InternalFolder,
    ActionSpace,
)

from .memory_metadata import MemoryMetadata, MemorySpace

from .swarm_history import SwarmHistory

from .swarm_state import SwarmState

from .swarm_nodes import (
    SwarmNode,
)

from .util_metadata import UtilSpace, UtilMetadata

from .swarm_operations import (
    SwarmOperation,
    SpawnOperation,
    TerminationOperation,
    FailureOperation,
    BlockingOperation,
    NodeEmbryo,
)

from .base_action import BaseAction


"""
This agent is called when all the children of a "DecomposeDirective" node have terminated.

It is crucial to note, the "DecomposeDirective" node breaks a directive into parallel 
subdirectives to be executed independently. There may still be further steps that need to be
taken following the completion of this set of subdirectives.

This agent will confirm the completion of the directive. If complete, it will terminate
and signal the parent node to terminate as well. Otherwise, it will spawn a new deompose
directive node to continue the process.
"""
from typing import List, Optional
from pydantic import BaseModel, Field

from swarmstar.swarm.types import BlockingOperation, SwarmState, TerminationOperation, SpawnOperation
from swarmstar.swarm.types.base_action import BaseAction


class ConfirmCompletionModel(BaseModel):
    is_complete: bool = Field(
        ..., description="Whether the directive is complete or not."
    )
    message: Optional[str] = Field(
        None, description="If the directive is not complete, this message will be sent back to the decompose directive node."
    )


CONFIRM_COMPLETION_INSTRUCTIONS = (
    "Your purpose is to review and determine if the directive has been completed. "
    "The node above you was given a directive, which it decomposed into subdirectives "
    "which were executed independently and in parallel. However, there may still be further "
    "steps that need to be taken following the completion of this set of subdirectives. "
    "You are given the directive and reports of all the child branches. Determine if the "
    "directive is complete or not. If it is not complete, generate a comprehensive message "
    "to send back to the decompose directive node, with all the information it might need to "
    "continue the process, to generate the next sequential set of subdirectives. The message "
    "you generate is all the node will see, so make sure it is comprehensive and actionable. "
)

class ConfirmCompletion(BaseAction):
    def main(self) -> BlockingOperation:
        '''
        Get reports of all child branches
        '''
        swarm_state = SwarmState(swarm=self.swarm)
        parent_node = swarm_state[self.node.parent_id]
        subdirectives = parent_node.journal[-1]["content"]
        initial_directive = parent_node.message
        child_branch_reports, branch_subdirectives = self.get_child_branch_reports()
        stringified_reports = ""
        for branch_report, branch_subdirective in zip(child_branch_reports, branch_subdirectives):
            joined_branch_reports = "\n".join(branch_report)
            stringified_reports += f"This branch's overarching directive is: {branch_subdirective}\n\n" \
                                   f"Here is the last journal entry of every node in this branch, stopping at a stop or leaf node:\n" \
                                   f"{joined_branch_reports}\n\n"
        
        messages = [
            {
                "role": "system", 
                "content": CONFIRM_COMPLETION_INSTRUCTIONS
            },
            {
                "role": "system",
                "content": (
                    f"The parent decomposed the directive: \n`{initial_directive}`\n\n",
                    f"Into subdirectives: {subdirectives}\n\n",
                    "Here are the reports of all the child branches:\n",
                    stringified_reports
                ),
            },
        ]
        
        self.add_journal_entry({
            "header": "Confirming Completion of Directive",
            "content": parent_node.message
        })
        
        return BlockingOperation(
            node_id=self.node.node_id,
            blocking_type="instructor_completion",
            args={
                "messages": messages,
                "instructor_model_name": "ConfirmCompletionModel",
            },
            context={},
            next_function_to_call="analyze_output",
        )


    def analyze_output(self, completion: ConfirmCompletionModel) -> BlockingOperation:
        '''
        Depending on the completion, we will either terminate or spawn a new decompose directive node.
        '''
        if completion.is_complete:
            self.add_journal_entry({
                "header": "Directive Complete",
                "content": "The directive has been completed."
            })
            return TerminationOperation(
                node_id=self.node.node_id,
            )
        else:
            self.add_journal_entry({
                "header": "Directive Incomplete",
                "content": f"The directive is incomplete. The message to send back to the decompose directive node is: {completion.message}"
            })
            return SpawnOperation(
                node_id=self.node.node_id,
                node_embryo={
                    "action_id": "swarmstar/actions/reasoning/decompose_directive",
                    "message": completion.message
                }
            )

    def get_child_branch_reports(self) -> List[List[str]]:
        '''
        This function will return a list of list of strings (all_branch_reports) 
        and a list of strings (branch_subdirectives)
        
        Each string in all_branch_reports contains the last journal entry of 
        a node in a branch.
        Each sublist in all_branch_reports represents a branch of the tree.
        
        Each string in branch_subdirectives contains the overarching directive
        of that entire branch.
        '''
        # When descending a branch, stop at these nodes
        # This should contain any nodes that have multiple children
        stop_action_ids = ["swarmstar/actions/reasoning/decompose_directive"]

        swarm_state = SwarmState(swarm=self.swarm)
        parent_node = swarm_state[self.node.parent_id]
        all_branch_reports = []
        branch_subdirectives = []
        for child_id in parent_node.children_ids: # Loop through all children of decompose directive node
            if child_id == self.node.node_id: # Skip this node (the confirm completion node)
                continue
            child_node = swarm_state[child_id]
            branch_subdirectives.append(child_node.message)
            this_branches_reports = []
            while True: # Descend down the branch, attaching the last journal entry of each node
                last_journal_entry = child_node.journal[-1]
                this_branches_reports.append(f"{last_journal_entry['header']}: {last_journal_entry['content']}")
                if child_node.action_id in stop_action_ids:  
                    break
                if child_node.children_ids:
                    child_id = child_node.children_ids[0]
                else:
                    break
            all_branch_reports.append(this_branches_reports)
        return all_branch_reports, branch_subdirectives



"""
This is a common interface to KV stores for the swarm space.

Things like action, util and memory space metadata, swarm state, 
swarm history and more will be stored in KV stores.
"""
from __future__ import annotations

from importlib import import_module
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from swarmstar.swarm.types.swarm_config import SwarmConfig


platform_map = {
    "mac": "swarmstar.utils.data.kv_operations.mongodb",
    "azure": "swarmstar.utils.data.kv_operations.cosmos_db",
}


def add_kv(swarm: SwarmConfig, category: str, key: str, value: dict) -> None:
    platform = swarm.platform
    kv_operations_module = import_module(platform_map[platform])
    return kv_operations_module.add_kv(swarm, category, key, value)


def get_kv(swarm: SwarmConfig, category: str, key: str) -> dict:
    platform = swarm.platform
    kv_operations_module = import_module(platform_map[platform])
    return kv_operations_module.get_kv(swarm, category, key)


def delete_kv(swarm: SwarmConfig, category: str, key: str) -> None:
    platform = swarm.platform
    kv_operations_module = import_module(platform_map[platform])
    return kv_operations_module.delete_swarm_space_kv_document(swarm, category, key)


def update_kv(swarm: SwarmConfig, category: str, key: str, value: dict) -> None:
    platform = swarm.platform
    kv_operations_module = import_module(platform_map[platform])
    return kv_operations_module.update_kv(swarm, category, key, value)



"""
The swarm history is a record of all the events that have occurred in the swarm.
It stores all the information needed to reconstruct the state of the swarm at any
point in time.
"""
from pydantic import BaseModel

from swarmstar.swarm.types.swarm_config import SwarmConfig
from swarmstar.swarm.types.swarm_operations import SwarmOperation
from swarmstar.utils.data.kv_operations.main import add_kv, get_kv


class SwarmHistory(BaseModel):
    swarm: SwarmConfig

    def __getitem__(self, frame: int) -> SwarmOperation:
        return SwarmOperation.model_validate(get_kv(self.swarm, "swarm_history", frame))

    def add_event(self, operation: SwarmOperation):
        current_frame = get_kv(self.swarm, "swarm_history", "current_frame")["frame"]
        add_kv(self.swarm, "swarm_history", current_frame, operation.model_dump())
        add_kv(
            self.swarm, "swarm_history", "current_frame", {"frame": current_frame + 1}
        )



"""
Throughout the conversation we maintain a "Conversation State" which is a data structure that contains the following:
    - A list of questions that need to be answered.
    - A concise and compact representation of only the necessary information to persist through the conversation.
    - A list of reports that is built up throughout the conversation. This report is what will be sent back when the conversation ends.
"""

from typing import List

from pydantic import BaseModel, Field

from swarmstar.swarm.types import BlockingOperation, TerminationOperation
from swarmstar.swarm.types.base_action import BaseAction


class QuestionAskerConversationState(BaseModel):
    questions: List[str] = Field(..., description="List of questions we need answered")
    persisted_context: str = Field(
        ...,
        description="A concise and compact representation of the necessary information to persist through the conversation.",
    )
    report: str = Field(
        ...,
        description="Concise and comprehensive report of answers to our questions and supporting context.",
    )


class AgentMessage(BaseModel):
    content: str = Field(..., description="The message to send to the user.")


class FinalReport(BaseModel):
    report: str = Field(
        ...,
        description="Concise and comprehensive report of answers to our questions and supporting context.",
    )


GENERATE_INITIAL_CONVERSATION_STATE_INSTRUCTIONS = (
    "Steps:\n\n"
    "1. Identify the questions that need to be answered.\n"
    "2. Persist critical context that is needed for the conversation from the goal. This "
    "context should be minimal and directly support the agent in maintaining the conversation.\n"
    "3. Craft an initial message to the user that is concise and clear to avoid confusion.\n\n"
    "Leave the report field empty. You are initializing the conversation state\n\n"
    "The reports aren't used as context throughout the conversation, so make sure all necessary context is in the persisted context."
).replace("\n", "\\n")

UPDATE_CONVERSATION_STATE_INSTRUCTIONS = (
    "Refine the conversation state based on the user's latest response.\n\n"
    "Steps:\n"
    "1. Update Questions: Rewrite the list of questions by removing resolved items and adding new ones as needed.\n"
    "2. Update Context: Rewrite the persisted context based off the most recent message. Keep it focused on essential "
    "information required for ongoing dialogue.\n"
    "3. Add to Reports: Add answers and relevant information to the reports given the users message. This report is what "
    "will be sent back when the conversation ends. Keep it brief and actionable yet comprehensive. What you say will be "
    "appended to the current reports so do not repeat stuff already mentioned.\n\n"
    "The reports aren't used as context throughout the conversation, so make sure all necessary context is in the persisted context."
).replace("\n", "\\n")

GENERATE_MESSAGE_INSTRUCTIONS = (
    "Generate a message to send to the user based on the current conversation state and the user's most recent message.\n"
    "You are given a list of questions that need to be answered. Don't stray off topic and aim to get answers to the questions."
).replace("\n", "\\n")

FINALIZE_REPORT_INSTRUCTIONS = (
    "Consolidate the list of reports into a final report."
).replace("\n", "\\n")


class AskUserQuestions(BaseAction):
    def main(self):
        return self.generate_initial_conversation_state()

    def generate_initial_conversation_state(self):
        messages = [
            {
                "role": "system",
                "content": GENERATE_INITIAL_CONVERSATION_STATE_INSTRUCTIONS,
            },
            {
                "role": "user",
                "content": f"You are initializing the conversation state. Extract questions and context from this message:\n`{self.node.message}`",
            },
        ]

        self.add_journal_entry({
                "header": "Asking User Questions",
                "content": self.node.message,   
        })

        return BlockingOperation(
            node_id=self.node.node_id,
            blocking_type="instructor_completion",
            args={
                "messages": messages,
                "instructor_model_name": "QuestionAskerConversationState",
            },
            context={
                "user_message": "User has no messages, the convo is just starting.",
                "recent_ai_message": "No messages yet, the convo is just starting.",
                "reports": [],
            },
            next_function_to_call="generate_message",
        )

    def generate_message(
        self,
        user_message: str,
        recent_ai_message: str,
        reports: List[str],
        completion: QuestionAskerConversationState,
    ):
        messages = [
            {"role": "system", "content": GENERATE_MESSAGE_INSTRUCTIONS},
            {
                "role": "system",
                "content": (
                    f"Questions: {completion.questions}\n\n"
                    f"Context: {completion.persisted_context}\n\n"
                    f"User's most recent message: {user_message}"
                    f"Your most recent message: {recent_ai_message}"
                ),
            },
        ]

        return BlockingOperation(
            node_id=self.node.node_id,
            blocking_type="instructor_completion",
            args={"messages": messages, "instructor_model_name": "AgentMessage"},
            context={
                "questions": completion.questions,
                "persisted_context": completion.persisted_context,
                "reports": reports,
            },
            next_function_to_call="send_user_message",
        )

    def send_user_message(
        self,
        questions: List[str],
        persisted_context: str,
        reports: List[str],
        completion: AgentMessage,
    ):
        return BlockingOperation(
            node_id=self.node.node_id,
            blocking_type="send_user_message",
            args={
                "message": completion.content,
            },
            context={
                "questions": questions,
                "persisted_context": persisted_context,
                "reports": reports,
                "recent_ai_message": completion.content,
            },
            next_function_to_call="update_conversation_state",
        )

    def update_conversation_state(
        self,
        questions: List[str],
        persisted_context: str,
        reports: List[str],
        recent_ai_message: str,
        user_response: str,
    ):
        messages = [
            {"role": "system", "content": UPDATE_CONVERSATION_STATE_INSTRUCTIONS},
            {
                "role": "user",
                "content": f"Your most recent message: {recent_ai_message}\n\nUsers most recent message: {user_response}",
            },
            {
                "role": "system",
                "content": f"Update Questions: {questions}\n\nUpdate Context: {persisted_context}\n\nAdd to Reports. Do not repeat things that are already mentioned: {reports}",
            },
        ]

        return BlockingOperation(
            node_id=self.node.node_id,
            blocking_type="instructor_completion",
            args={
                "messages": messages,
                "instructor_model_name": "QuestionAskerConversationState",
            },
            context={
                "reports": reports,
                "user_response": user_response,
                "recent_ai_message": recent_ai_message,
            },
            next_function_to_call="decide_to_continue_or_end_conversation",
        )

    def decide_to_continue_or_end_conversation(
        self,
        reports: List[str],
        user_response: str,
        recent_ai_message: str,
        completion: QuestionAskerConversationState,
    ):
        reports.append(completion.report)
        if (not completion.questions) or (len(completion.questions) == 0):
            return self.finalize_report(reports)

        return self.generate_message(
            user_response, recent_ai_message, reports, completion
        )

    def finalize_report(self, reports: List[str]):
        messages = [
            {"role": "system", "content": FINALIZE_REPORT_INSTRUCTIONS},
            {"role": "system", "content": f"Reports: {reports}"},
        ]

        return BlockingOperation(
            node_id=self.node.node_id,
            blocking_type="instructor_completion",
            args={"messages": messages, "instructor_model_name": "FinalReport"},
            context={},
            next_function_to_call="terminate_conversation",
        )

    def terminate_conversation(self, completion: FinalReport):
        self.add_journal_entry(
            {
                "header": "Final Report",
                "content": completion.report,
            }
        )
        return TerminationOperation(node_id=self.node.node_id)



"""
The NodeEmbryo is what a node outputs to spawn children.

Nodes can perform 1 of 4 "SwarmOperations":
    - SpawnOperation
    - ExecuteOperation
    - TerminationOperation
    - FailureOperation
    - BlockingOperation
"""
from typing import Any, Dict, Literal, Optional

from pydantic import BaseModel, Field


class NodeEmbryo(BaseModel):
    action_id: str
    message: str


class SwarmOperation(BaseModel):
    operation_type: Literal["spawn", "terminate", "node_failure", "blocking"]
    node_id: str


class BlockingOperation(SwarmOperation):
    operation_type: Literal["blocking"] = Field(default="blocking")
    node_id: str
    blocking_type: str
    args: Dict[str, Any] = {}
    context: Dict[str, Any] = {}
    next_function_to_call: str

class SpawnOperation(SwarmOperation):
    operation_type: Literal["spawn"] = Field(default="spawn")
    node_embryo: NodeEmbryo
    termination_policy_change: Literal[
        "simple", "parallel_review", "clone_with_questions_answered"
    ] = None
    node_id: str = None


class TerminationOperation(SwarmOperation):
    operation_type: Literal["terminate"] = Field(default="terminate")
    node_id: str

class FailureOperation(SwarmOperation):
    operation_type: Literal["node_failure"] = Field(default="node_failure")
    node_id: Optional[str] = None
